import astropy.units as u

from ...core.math import Random
from .weather_station_base import WeatherStation


class WeatherStationSimulator(WeatherStation):

    Manufacturer: str = ""
    Model: str = ""
    Identifier = ""
    is_simulator = True

    def __init__(self) -> None:
        self._random_temp = Random().walk(15, 0.1, -10)
        self._random_hum = Random(limits=(0, 1)).walk(0.3, 0.01, -7)
        self._random_pres = Random().walk(850, 0.1, -7)

    def get_temperature(self) -> u.Quantity:
        return next(self._random_temp) * u.deg_C

    def get_humidity(self) -> float:
        return next(self._random_hum)

    def get_pressure(self) -> u.Quantity:
        return next(self._random_pres) * u.hPa  # type: ignore

    def finalize(self) -> None:
        pass



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple, Union

import astropy.units as u
import numpy as np

from ...core.normalization import get_quantity
from ...core.types import CoordFrameType, DimensionLess, UnitType
from .path_base import ControlContext, Index, Path

if TYPE_CHECKING:
    from ..convert import CoordCalculator

T = Union[DimensionLess, u.Quantity]


class Linear(Path):

    tight = True
    infinite = False
    waypoint = False

    def __init__(
        self,
        calc: CoordCalculator,
        *target: Union[str, DimensionLess, u.Quantity, CoordFrameType],
        unit: Optional[UnitType] = None,
        start: Tuple[T, T],
        stop: Tuple[T, T],
        scan_frame: CoordFrameType,
        speed: T,
        offset: Optional[Tuple[T, T, CoordFrameType]] = None,
        margin: Optional[T] = None,
        **ctx_kw: Any,
    ) -> None:
        super().__init__(calc, *target, unit=unit)

        self._start = get_quantity(start, unit=unit)  # type: ignore
        self._stop = get_quantity(stop, unit=unit)  # type: ignore
        self._scan_frame = scan_frame
        self._speed = get_quantity(speed, unit=f"{unit}/s")
        self._offset = (
            None
            if offset is None
            else calc.coordinate_delta(
                d_lon=offset[0], d_lat=offset[1], frame=offset[2], unit=unit
            )
        )
        self._ctx_kw = ctx_kw

        require_unit = (
            ((offset is not None) and (not isinstance(offset[0], u.Quantity)))
            or ((len(target) > 0) and (not isinstance(target[0], u.Quantity)))
            or (not isinstance(start[0], u.Quantity))
            or (not isinstance(stop[0], u.Quantity))
        )
        if (unit is None) and require_unit:
            raise ValueError(
                "Unit must be specified if any of the following arguments are not "
                "quantities; `target`, `start`, `stop`, `offset`"
            )
        self._unit = unit

        if margin is None:
            raise NotImplementedError("Auto-margin is not implemented yet")
        self._margin = get_quantity(margin, unit=unit)

    @property
    def n_cmd(self) -> Union[int, float]:
        distance: u.Quantity = np.linalg.norm(
            np.asanyarray(self._start) - np.asanyarray(self._stop)  # type: ignore
        )
        duration_sec = (distance / self._speed).to_value("s")
        return float(duration_sec * self._calc.command_freq)  # type: ignore

    @property
    def target_frame(self) -> CoordFrameType:
        if self._offset is not None:
            return self._offset.frame
        return self._scan_frame

    @property
    def lonlat_func(self) -> Callable[[Index], Tuple[T, T]]:
        def _lonlat_func(idx: Index) -> Tuple[T, T]:
            if self._target is None:
                kw = dict(frame=self._scan_frame, unit=self._unit, time=idx.time)
                _start = self._calc.coordinate(
                    lon=self._start[0], lat=self._start[1], **kw
                )
                _stop = self._calc.coordinate(
                    lon=self._stop[0], lat=self._stop[1], **kw
                )
            else:
                _reference = (
                    self._target.realize(time=idx.time)  # type: ignore
                    if hasattr(self._target, "realize")
                    else self._target
                )
                _reference = _reference.replicate(time=idx.time)
                offset_to_start = self._calc.coordinate_delta(
                    d_lon=self._start[0], d_lat=self._start[1], frame=self._scan_frame
                )
                offset_to_stop = self._calc.coordinate_delta(
                    d_lon=self._stop[0], d_lat=self._stop[1], frame=self._scan_frame
                )
                _start = _reference.cartesian_offset_by(offset_to_start)
                _stop = _reference.cartesian_offset_by(offset_to_stop)

            start = (
                _start
                if self._offset is None
                else _start.cartesian_offset_by(self._offset)
            )
            stop = (
                _stop
                if self._offset is None
                else _stop.cartesian_offset_by(self._offset)
            )

            _ratio = idx.index / self.n_cmd  # type: ignore
            lon = start.lon * (1 - _ratio) + stop.lon * _ratio
            lat = start.lat * (1 - _ratio) + stop.lat * _ratio

            return lon, lat

        return _lonlat_func

    @property
    def arguments(self) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        ctx_kw = dict(
            tight=self.tight,
            infinite=self.infinite,
            waypoint=self.waypoint,
            duration=self.n_cmd / self._calc.command_freq,
        )
        ctx_kw.update(self._ctx_kw)

        kwargs: Dict[str, Any] = dict(
            n_cmd=self.n_cmd, context=ControlContext(**ctx_kw)
        )
        if self._unit is not None:
            kwargs.update(unit=self._unit)

        return (self.lonlat_func, self.target_frame), {
            "n_cmd": self.n_cmd,
            "context": ControlContext(**ctx_kw),
            "unit": self._unit,
        }


class Accelerate(Linear):

    tight = False
    infinite = False
    waypoint = False

    @property
    def n_cmd(self) -> Union[int, float]:
        a = (self._speed**2) / (2 * self._margin)
        duration = ((2 * self._margin) / a) ** (1 / 2)
        return float(duration.to_value("s") * self._calc.command_freq)  # type: ignore

    @property
    def lonlat_func(self) -> Callable[[Index], Tuple[T, T]]:
        scan_vector: u.Quantity = self._stop - self._start  # type: ignore
        unit_scan_vector = scan_vector / np.linalg.norm(scan_vector) * scan_vector.unit
        margin_start = self._start - unit_scan_vector * self._margin.to_value(
            unit_scan_vector.unit
        )

        def _lonlat_func(idx: Index) -> Tuple[T, T]:
            if self._target is None:
                kw = dict(frame=self._scan_frame, unit=self._unit, time=idx.time)
                _start = self._calc.coordinate(
                    lon=margin_start[0], lat=margin_start[1], **kw
                )
                _stop = self._calc.coordinate(
                    lon=self._start[0], lat=self._start[1], **kw
                )
            else:
                kw = dict(obstime=idx.time, unit=self._unit)
                _reference = (
                    self._target.realize(time=idx.time)  # type: ignore
                    if hasattr(self._target, "realize")
                    else self._target
                )
                _reference = _reference.replicate(time=idx.time)
                offset_to_start = self._calc.coordinate_delta(
                    d_lon=margin_start[0], d_lat=margin_start[1], frame=self._scan_frame
                )
                offset_to_stop = self._calc.coordinate_delta(
                    d_lon=self._start[0], d_lat=self._start[1], frame=self._scan_frame
                )
                _start = _reference.cartesian_offset_by(offset_to_start)
                _stop = _reference.cartesian_offset_by(offset_to_stop)

            start = (
                _start
                if self._offset is None
                else _start.cartesian_offset_by(self._offset)
            )
            stop = (
                _stop
                if self._offset is None
                else _stop.cartesian_offset_by(self._offset)
            )

            _ratio = idx.index / self.n_cmd  # type: ignore
            lon = start.lon * (1 - _ratio**2) + stop.lon * _ratio**2
            lat = start.lat * (1 - _ratio**2) + stop.lat * _ratio**2

            return lon, lat

        return _lonlat_func

    @property
    def arguments(self) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        ctx_kw = dict(
            tight=self.tight,
            infinite=self.infinite,
            waypoint=self.waypoint,
            duration=self.n_cmd / self._calc.command_freq,
        )
        ctx_kw.update(self._ctx_kw)

        kwargs: Dict[str, Any] = dict(
            n_cmd=self.n_cmd, context=ControlContext(**ctx_kw)
        )
        if self._unit is not None:
            kwargs.update(unit=self._unit)

        return (self.lonlat_func, self.target_frame), {
            "n_cmd": self.n_cmd,
            "context": ControlContext(**ctx_kw),
            "unit": self._unit,
        }


class Standby(Linear):

    tight = True
    infinite = True
    waypoint = True

    @property
    def n_cmd(self) -> Union[int, float]:
        return self._calc.command_freq * self._calc.command_group_duration_sec

    @property
    def lonlat_func(self) -> Callable[[Index], Tuple[T, T]]:
        scan_vector: u.Quantity = self._stop - self._start  # type: ignore

        unit_scan_vector = scan_vector / np.linalg.norm(scan_vector) * scan_vector.unit
        margin_start = self._start - unit_scan_vector * self._margin.to_value(
            unit_scan_vector.unit
        )

        def _lonlat_func(idx: Index) -> Tuple[T, T]:
            if self._target is None:
                kw = dict(frame=self._scan_frame, unit=self._unit, time=idx.time)
                _start = self._calc.coordinate(
                    lon=margin_start[0], lat=margin_start[1], **kw
                )
            else:
                kw = dict(obstime=idx.time, unit=self._unit)
                _reference = (
                    self._target.realize(time=idx.time)  # type: ignore
                    if hasattr(self._target, "realize")
                    else self._target
                )
                _reference = _reference.replicate(time=idx.time)
                offset = self._calc.coordinate_delta(
                    d_lon=margin_start[0], d_lat=margin_start[1], frame=self._scan_frame
                )
                _start = _reference.cartesian_offset_by(offset)

            start = (
                _start
                if self._offset is None
                else _start.cartesian_offset_by(self._offset)
            )
            lon = np.broadcast_to(start.lon, idx.time.shape)  # type: ignore
            lat = np.broadcast_to(start.lat, idx.time.shape)  # type: ignore

            return lon, lat

        return _lonlat_func

    @property
    def arguments(self) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        ctx_kw = dict(
            tight=self.tight,
            infinite=self.infinite,
            waypoint=self.waypoint,
            duration=self._calc.command_group_duration_sec,
        )
        ctx_kw.update(self._ctx_kw)

        kwargs: Dict[str, Any] = dict(
            n_cmd=self.n_cmd, context=ControlContext(**ctx_kw)
        )
        if self._unit is not None:
            kwargs.update(unit=self._unit)

        return (self.lonlat_func, self.target_frame), {
            "n_cmd": self.n_cmd,
            "context": ControlContext(**ctx_kw),
            "unit": self._unit,
        }



"""String or rich-text representation of NECLIB data types."""

__all__ = ["html_repr_of_dict"]

import re
from collections import defaultdict
from typing import Any, Dict, List, Optional, Type


def html_repr_of_dict(
    __dict: Dict[Any, Any],
    __type: Optional[Type[Any]] = None,
    /,
    *,
    aliases: Optional[Dict[str, str]] = None,
    metadata: Optional[Dict[str, Any]] = None,
) -> str:
    """Return a HTML representation of a dictionary.

    Parameters
    ----------
    __dict
        Dictionary to be represented.
    __type
        Type of the object, which this dictionary represents.
    metadata
        Metadata to be included in the representation.

    Returns
    -------
    rich_repr
        The HTML representation of the dictionary.

    """
    if __type is None:
        __type = type(__dict)
    if aliases is None:
        aliases = {}
    if metadata is None:
        metadata = {}

    pkg = getattr(__type, "__module__", "")
    typename = getattr(__type, "__name__", str(__type))
    type_repr = f"<span>{f'{pkg}.' if pkg else ''}{typename}</span><hr>"

    element_repr = [
        f"<tr><td>{k}</td><td><code>{v}</code></td><td>{type(v).__name__}</td></tr>"
        for k, v in __dict.items()
    ]

    metadata_repr = [
        f"<tr><td>{k}</td><td><code>{v}</code></td><td>{type(v).__name__}</td></tr>"
        for k, v in metadata.items()
    ]

    _aliases: Dict[str, List[str]] = defaultdict(list)
    # Reverse the aliases to show [Actual Key : Alias Key 1, Alias Key 2] style
    # representation
    for k, v in aliases.items():
        _aliases[v].append(k)
    alias_repr = [
        f"<tr><td><code>{k}</code></td><td><code>{', '.join(v)}</code></td></tr>"
        for k, v in _aliases.items()
    ]

    repr_ = f"""
    {type_repr}
    <details><summary>{len(element_repr)} elements</summary>
        <table>
            <thead>
                <tr>
                    <th>Key</th>
                    <th>Value</th>
                    <th>Type</th>
                </tr>
            </thead>
            <tbody>
                {''.join(element_repr)}
            </tbody>
        </table>
    </details>
    <details><summary>{len(alias_repr)} elements have alias key(s)</summary>
        <table>
            <thead>
                <tr>
                    <th>Key</th>
                    <th>Aliases</th>
                </tr>
            </thead>
            <tbody>
                {''.join(alias_repr)}
            </tbody>
        </table>
    </details>
    <details><summary>{len(metadata_repr)} metadata</summary>
        <table>
            <thead>
                <tr>
                    <th>Key</th>
                    <th>Value</th>
                    <th>Type</th>
                </tr>
            </thead>
            <tbody>
                {''.join(metadata_repr)}
            </tbody>
        </table>
    </details>
    """
    return re.sub(r"\n\s*", "", repr_)  # Remove all newlines and indentation



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple, Union

import astropy.units as u

from ...core.types import CoordFrameType, DimensionLess, UnitType
from .path_base import ControlContext, Index, Path

if TYPE_CHECKING:
    from ..convert import CoordCalculator

T = Union[DimensionLess, u.Quantity]


class Track(Path):

    tight = True
    infinite = True
    waypoint = False

    def __init__(
        self,
        calc: CoordCalculator,
        *target: Union[str, DimensionLess, u.Quantity, CoordFrameType],
        unit: Optional[UnitType] = None,
        offset: Optional[Tuple[T, T, CoordFrameType]] = None,
        **ctx_kw: Any,
    ) -> None:
        super().__init__(calc, *target, unit=unit)

        self._offset = (
            None
            if offset is None
            else calc.coordinate_delta(
                d_lon=offset[0], d_lat=offset[1], frame=offset[2], unit=unit
            )
        )
        self._ctx_kw = ctx_kw

        require_unit = (
            (offset is not None) and (not isinstance(offset[0], u.Quantity))
        ) or (not isinstance(target[0], (str, u.Quantity)))
        if (unit is None) and require_unit:
            raise ValueError(
                "`unit` is required when `target` or `offset` is given as float value."
            )
        self._unit = unit

    @property
    def n_cmd(self) -> Union[int, float]:
        return self._calc.command_freq * self._calc.command_group_duration_sec

    @property
    def lonlat_func(self) -> Callable[[Index], Tuple[T, T]]:
        def _lonlat_func(idx: Index) -> Tuple[T, T]:
            target = (
                self._target.realize(time=idx.time)  # type: ignore
                if hasattr(self._target, "realize")
                else self._target
            )
            target = target.replicate(time=idx.time)  # type: ignore
            offset_applied = (
                target.cartesian_offset_by(self._offset) if self._offset else target
            )
            return offset_applied.lon, offset_applied.lat

        return _lonlat_func

    @property
    def arguments(self) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        ctx_kw = dict(
            tight=self.tight,
            infinite=self.infinite,
            waypoint=self.waypoint,
            duration=self._calc.command_group_duration_sec,
        )
        ctx_kw.update(self._ctx_kw)

        kwargs: Dict[str, Any] = dict(
            n_cmd=self.n_cmd, context=ControlContext(**ctx_kw)
        )
        if self._unit is not None:
            kwargs.update(unit=self._unit)

        return (self.lonlat_func, self.target_frame), {
            "n_cmd": self.n_cmd,
            "context": ControlContext(**ctx_kw),
            "unit": self._unit,
        }



from typing import Tuple

import astropy.units as u
import numpy as np

from .pointing_error import PointingError


class NANTEN2(PointingError):
    r"""Pointing model used in NANTEN2.

    .. math::

        \Delta x =& \chi_{Az} \sin ( \omega_{Az} - Az ) \sin ( El ) \\
        &+ \epsilon \sin ( El ) \\
        &+ \chi_{2, Az} \sin ( 2 ( \omega_{2, Az} - Az ) ) \sin ( El ) \\
        &+ \mathrm{d}Az \cos ( El ) \\
        &+ \mathrm{d}e \\
        &+ cor_\mathrm{v} \cos ( El + cor_\mathrm{p} ) \\
        &+ \mathrm{d}e_\mathrm{radio} \\
        \Delta Az =& \Delta x / \cos ( El ) \\
        \Delta y =& - \chi_{El} \cos ( \omega_{El} - Az ) \\
        &- \chi_{2, El} \cos ( 2 ( \omega_{2, El} - Az ) ) \\
        &+ g_1 \cos ( El ) + g_2 \sin ( El ) \\
        &+ \mathrm{d}el \\
        &+ g_{ 1,\mathrm{radio} } \cos ( El ) + g_{ 2,\mathrm{radio} } \sin ( El ) \\
        &- cor_\mathrm{v} \sin ( El + cor_\mathrm{p} ) \\
        &+ \mathrm{d}el_\mathrm{radio} \\
        \Delta El =& \Delta y

    Parameters
    ----------
    dAz
        Azimuth (not X) offset of encoder reading.
    de
        X collimation error.
    chi_Az
        Magnitude of tilt of azimuth axis.
    omega_Az
        Phase (azimuth direction) of tilt of azimuth axis.
    eps
        Skew angle (lack of orthogonality) between azimuth and elevation axes.
    chi2_Az
        Same as chi, but the period is 180deg in azimuth axis (harmonic component).
    omega2_Az
        Same as omega, but the period is 180deg in azimuth (harmonic component).
    chi_El
        Magnitude of tilt of elevation axis.
    omega_El
        Phase (azimuth direction) of tilt of elevation axis.
    chi2_El
        Same as chi, but the period is 180deg in azimuth axis (harmonic component).
    omega2_El
        Same as omega, but the period is 180deg in azimuth (harmonic component).
    g
        First order gravitational deflection coefficient.
    gg
        Second order gravitational deflection coefficient.
    ggg
        Third order gravitational deflection coefficient.
    gggg
        Fourth order gravitational deflection coefficient.
    dEl: Angle = 0 << u.arcsec
        Elevation offset of encoder reading.
    de_radio: Angle = 0 << u.arcsec
        Constant X (not azimuth) offset between optical and radio beam.
    del_radio: Angle = 0 << u.arcsec
        Constant elevation offset between optical and radio beam.
    cor_v: Angle = 0 << u.arcsec
        Amplitude of collimation error.
    cor_p: Angle = 0 << u.deg
        Phase of collimation error, negative of elevation where the elevation component
        of collimation error is zero.
    g_radio: float = 0
        First order gravitational deflection coefficient.
    gg_radio: float = 0
        Second order gravitational deflection coefficient.
    ggg_radio: float = 0
        Third order gravitational deflection coefficient.
    gggg_radio: float = 0
        Fourth order gravitational deflection coefficient.

    """

    def offset(self, az: u.Quantity, el: u.Quantity) -> Tuple[u.Quantity, u.Quantity]:
        gravitational_term = np.polynomial.Polynomial(
            [0, self.g, self.gg, self.ggg, self.gggg]
        )
        radio_gravitational_term = np.polynomial.Polynomial(
            [0, self.g_radio, self.gg_radio, self.ggg_radio, self.gggg_radio]
        )

        dx = (
            self.chi_Az * np.sin(self.omega_Az - az) * np.sin(el)
            + self.eps * np.sin(el)
            + self.chi2_Az * np.sin(2 * (self.omega2_Az - az)) * np.sin(el)
            + self.dAz * np.cos(el)
            + self.de
            + self.cor_v * np.cos(el + self.cor_p)
            + self.de_radio
        )  # NOTE: 3rd term: chi2 sin sin, revised from chi2 cos sin
        dAz = dx / np.cos(el)
        dEl = (
            -1 * self.chi_El * np.cos(self.omega_El - az)
            - self.chi2_El * np.cos(2 * (self.omega2_El - az))
            + gravitational_term(el.to_value("deg")) * u.arcsec
            + self.dEl
            + radio_gravitational_term(el.to_value("deg")) * u.arcsec
            - self.cor_v * np.sin(el + self.cor_p)
            + self.del_radio
        )  # NOTE: 2nd term: chi2 cos, revised from chi2 sin
        # TODO: Review dimension of gravitational terms
        return dAz, dEl

    def fit(self, *args, **kwargs):
        raise NotImplementedError("Fitting is not implemented for this model.")



from .console_logger import get_logger
from .deprecation import deprecated, deprecated_namespace
from .disable import disabled

__all__ = ["deprecated", "deprecated_namespace", "get_logger", "disabled"]



from collections import defaultdict

import astropy.units as u

from ...core.units import dBm
from .attenuator_base import Attenuator


class AttenuatorSimulator(Attenuator):
    Manufacturer: str = ""
    Model: str = ""
    Identifier = ""
    is_simulator = True

    def __init__(self) -> None:
        self.loss = defaultdict(lambda: 0)

    def set_loss(self, dB: int, id: str):
        self.loss[id] = dB

    def get_loss(self, id: str) -> u.Quantity:
        return self.loss[id] * dBm

    def finalize(self) -> None:
        pass



import logging
import time
from typing import Dict, Optional, Union

from .. import environ


class ColorizeLevelNameFormatter(logging.Formatter):
    """Colorize severity level name.

    See `logging docs <https://docs.python.org/3/library/logging.html#handler-objects>`_
    for the usage.

    """

    ColorPrefix: Dict[int, str] = {
        0: "\x1b[0m",  # NOTSET, default (Black or White)
        10: "\x1b[35m",  # DEBUG, Magenta
        20: "\x1b[32m",  # INFO, Green
        30: "\x1b[33m",  # WARNING, Yellow
        40: "\x1b[31m",  # ERROR, Red
        50: "\x1b[41;97m",  # CRITICAL, White on Red
    }

    def format(self, record: logging.LogRecord) -> str:
        """Format a record to text."""
        levelno = int(min(max(0, record.levelno // 10 * 10), 50))
        original_levelname = record.levelname
        record.levelname = self.ColorPrefix[levelno] + original_levelname + "\x1b[0m"
        return super().format(record)


class Throttle(logging.Filter):
    """Reduce logging frequency of identical messages.

    Parameters
    ----------
    duration_sec
        Duration in seconds. If the same message is logged within this duration,
        attached handlers discard it.

    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, duration_sec: Union[int, float]):
        super().__init__(name=self.__class__.__name__)
        self._duration_sec = duration_sec
        if not hasattr(self, "_last_log_time"):
            # Calling singleton class will return the same instance, but the __init__
            # method will be called again. That will empty the time keeper, so
            # initialize this dictionary only when it is not defined.
            self._last_log_time: Dict[str, float] = {}

    def filter(self, record: logging.LogRecord) -> bool:
        """Filter function to be attached to logger."""
        now = time.time()
        for k in list(self._last_log_time.keys()):
            # Comparing `now` with `record.created` is not accurate, but this is to
            # reduce memory consumption, i.e., otherwise self._last_log_time would be
            # filled by infrequent logs.
            if now - self._last_log_time[k] > self._duration_sec:
                self._last_log_time.pop(k, None)

        key = (record.levelno, record.msg)
        if key in self._last_log_time:
            return False
        self._last_log_time[key] = record.created
        return True


def get_logger(
    name: Optional[str] = None,
    min_level: int = None,
    throttle_duration_sec: Union[int, float] = 1.0,
) -> logging.Logger:
    """Get logger instance which prints operation logs to console.

    Parameters
    ----------
    name
        Name of the logger. Calling this function with same ``name`` returns the same
        logger instance.
    min_level
        Lower bound of severity level to be displayed on terminal. To suppress less
        severe messages, set higher value. No matter this value, the log file contains
        all messages severer than ``logging.INFO`` (level=20).
    throttle_duration_sec
        Duration in seconds to throttle messages. If the same message is logged within
        this duration, the message is not displayed on terminal.

    Examples
    --------
    >>> logger = neclib.interfaces.getLogger("OTF_observation", "path/to/log.txt")
    >>> logger.debug("Inform something only needed on diagnosing some problem.")
    >>> logger.info("Inform something EXPECTED has happened.")
    >>> logger.warning("Inform something user should care has happened.")
    >>> logger.error(f"Notify some {functionality} cannot be completed due to error.")
    >>> logger.critical(
    ...     "Notify the %s cannot continue operation due to %s", "program", "some error"
    ... )
    >>> logger.obslog("Observation finished", indent_level=1)

    """
    logger_name = "neclib" if name is None else f"neclib.{name.strip('neclib.')}"
    logger = logging.getLogger("necst." + logger_name)
    [logger.removeHandler(f) for f in logger.filters if isinstance(f, Throttle)]
    logger.addFilter(Throttle(throttle_duration_sec))

    min_level = environ.log_level.get() if min_level is None else min_level

    fmt = "%(asctime)-s: [%(levelname)-s: %(filename)s#L%(lineno)s] %(message)s"
    color_log_format = ColorizeLevelNameFormatter(fmt)

    ch = logging.StreamHandler()
    ch.setLevel(min_level)
    ch.setFormatter(color_log_format)

    rootLogger = logging.getLogger()
    chs = [_ch for _ch in rootLogger.handlers if isinstance(_ch, logging.StreamHandler)]
    [rootLogger.handlers.remove(_ch) for _ch in chs]
    rootLogger.addHandler(ch)

    return logger



from typing import Generator

import astropy.units as u
import numpy as np

from .observation_spec_base import (
    ObservationMode,
    ObservationSpec,
    TimeKeeper,
    Waypoint,
)


class OTFSpec(ObservationSpec):
    __slots__ = ("_hot_time_keeper", "_off_time_keeper")

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self._hot_time_keeper = TimeKeeper(self["load_interval"])
        self._off_time_keeper = TimeKeeper(self["off_interval"])

    def observe(self) -> Generator[Waypoint, None, None]:
        for i, coord in enumerate(self._scan()):
            if self._hot_time_keeper.should_observe:
                self._hot_time_keeper.tell_observed()
                yield self.hot(f"{i}")

            if self._off_time_keeper.should_observe:
                self._off_time_keeper.tell_observed()
                yield self.off(f"{i}")

            self._hot_time_keeper.increment("scan")
            self._off_time_keeper.increment("scan")
            coord.id = f"{i}"
            yield coord

        # Should be executed at the end of the observation, to enable interpolation.
        yield self.hot("9999")
        yield self.off("9999")

    def _scan(self) -> Generator[Waypoint, None, None]:
        scan_length = self["scan_length"] * self["scan_velocity"]
        pa = self["position_angle"]
        _start_x, _start_y = self["start_position_x"], self["start_position_y"]
        start_x = _start_x * np.cos(pa) - _start_y * np.sin(pa)
        start_y = _start_x * np.sin(pa) + _start_y * np.cos(pa)
        for idx in range(int(self["first_scan"]), int(self["n"])):
            if self["scan_direction"].upper() == "X":
                offset = (0 * u.deg, idx * self["scan_spacing"])
            else:
                offset = (idx * self["scan_spacing"], 0 * u.deg)

            # Position angle correction
            offset = (
                offset[0] * np.cos(pa) - offset[1] * np.sin(pa),
                offset[0] * np.sin(pa) + offset[1] * np.cos(pa),
            )

            start = (start_x + offset[0], start_y + offset[1])
            stop = (
                start[0] + scan_length * np.cos(pa),
                start[1] + scan_length * np.sin(pa),
            )

            yield Waypoint(
                mode=ObservationMode.ON,
                reference=self._reference,
                start=start,
                stop=stop,
                speed=abs(self["scan_velocity"]),
                scan_frame=self["coord_sys"],
            )



"""Coordinate conversion functions.

This module contains functions for coordinate conversion and types for coordinate
handling. The intention of the custom type definitions is to avoid unintended coordinate
conversions which can be caused by the use of highly automated ``SkyCoord`` objects.

"""

__all__ = ["CoordinateDelta", "CoordCalculator", "CoordinateLike"]

import os
from dataclasses import dataclass, fields
from typing import Any, ClassVar, Dict, Optional, Tuple, Type, TypeVar, Union, overload

import astropy.units as u
import numpy as np
from astropy.coordinates import BaseCoordinateFrame, EarthLocation, SkyCoord, get_body
from astropy.coordinates.erfa_astrom import ErfaAstromInterpolator, erfa_astrom
from astropy.time import Time

from ..core import config, get_logger
from ..core.normalization import QuantityValidator, get_quantity
from ..core.types import Array, CoordFrameType, DimensionLess, UnitType
from .frame import parse_frame
from .pointing_error import PointingError

CoordinateLike = TypeVar("CoordinateLike", SkyCoord, BaseCoordinateFrame)
logger = get_logger(__name__, throttle_duration_sec=10)


class to_astropy_type:
    @overload
    @staticmethod
    def time(
        time: Union[str, int, float, Array[str], Time, Array[Union[int, float]]], /
    ) -> Time:
        ...

    @overload
    @staticmethod
    def time(time: None, /) -> None:
        ...

    @overload
    @staticmethod
    def time(
        *times: Union[str, int, float, Array[str], Time, Array[Union[int, float]]]
    ) -> Tuple[Time, ...]:
        ...

    @staticmethod
    def time(
        *times: Optional[
            Union[str, int, float, Time, Array[Union[int, float]], Array[str]]
        ]
    ) -> Optional[Union[Time, Tuple[Time, ...]]]:
        if (len(times) == 0) or (times[0] is None):
            return

        ret = []
        for time in times:
            if isinstance(time, Time):
                ret.append(time)
            elif isinstance(time, str):
                ret.append(Time(time))
            else:
                ret.append(Time(time, format="unix"))
        return tuple(ret) if len(ret) > 1 else ret[0]

    @overload
    @staticmethod
    def frame(
        frame: Union[str, BaseCoordinateFrame, Type[BaseCoordinateFrame]], /
    ) -> BaseCoordinateFrame:
        ...

    @overload
    @staticmethod
    def frame(frame: None, /) -> None:
        ...

    @overload
    @staticmethod
    def frame(
        *frames: Union[str, BaseCoordinateFrame, Type[BaseCoordinateFrame]]
    ) -> Tuple[BaseCoordinateFrame, ...]:
        ...

    @staticmethod
    def frame(
        *frames: Optional[Union[str, BaseCoordinateFrame, Type[BaseCoordinateFrame]]]
    ) -> Optional[Union[BaseCoordinateFrame, Tuple[BaseCoordinateFrame, ...]]]:
        if (len(frames) == 0) or (frames[0] is None):
            return

        ret = []
        for frame in frames:
            if isinstance(frame, BaseCoordinateFrame):
                ret.append(frame)
            elif isinstance(frame, str):
                ret.append(parse_frame(frame))
            elif callable(frame) and issubclass(frame, BaseCoordinateFrame):
                ret.append(frame())
            else:
                raise TypeError(f"Got invalid frame of type {type(frame)}: {frame}")
        return tuple(ret) if len(ret) > 1 else ret[0]


@dataclass
class Coordinate:
    lon: Union[DimensionLess, u.Quantity]
    lat: Union[DimensionLess, u.Quantity]
    frame: Union[Type[BaseCoordinateFrame], BaseCoordinateFrame, str]
    distance: Optional[Union[DimensionLess, u.Quantity]] = None
    time: Optional[Union[DimensionLess, Time]] = None
    unit: Optional[UnitType] = None

    _calc: ClassVar["CoordCalculator"]

    def __post_init__(self) -> None:
        if (self.lon is None) or (self.lat is None) or (self.frame is None):
            raise TypeError("Either `name` or `lon`, `lat` and `frame` must be given.")

        self.lon = get_quantity(self.lon, unit=self.unit)
        self.lat = get_quantity(self.lat, unit=self.unit)
        if self.distance is not None:
            # FIXME: invalid unit
            self.distance = get_quantity(self.distance, unit=self.unit)
        self.time = to_astropy_type.time(self.time)
        self.frame = to_astropy_type.frame(self.frame)

    @property
    def skycoord(self) -> SkyCoord:
        """SkyCoord object that represents this coordinate."""
        frame = self._normalize_frame(self.frame)
        kwargs = dict(obstime=self.time, **self._calc.altaz_kwargs)

        # To avoid the following error, attributes of current frame (self.frame)
        # shouldn't be passed as kwargs. Currently the filtering is applied implicitly
        # but this can lead to unexpected behavior, so this may emit warning in the
        # future.
        # ```Cannot specify frame attribute 'obstime' directly as an argument to
        # SkyCoord because a frame instance was passed in. Either pass a frame class,
        # or modify the frame attributes of the input frame instance.````
        _ = [kwargs.pop(k, None) for k in frame.frame_attributes]

        broadcasted = self.broadcasted
        args = [broadcasted.lon, broadcasted.lat]
        if self.distance is not None:
            args.append(broadcasted.distance)  # type: ignore
        return SkyCoord(*args, frame=frame, **kwargs)

    def _normalize_frame(
        self, frame: Union[Type[BaseCoordinateFrame], BaseCoordinateFrame]
    ) -> BaseCoordinateFrame:
        frame_kwargs = {}

        # Time dependent coordinates. Maybe more, but FK4 shouldn't be included here.
        if frame.name in ("altaz", "gcrs"):  # type: ignore
            frame_kwargs.update(obstime=self.time)

        if frame.name in ("altaz",):  # type: ignore
            frame_kwargs.update(self._calc.altaz_kwargs)

        if isinstance(frame, BaseCoordinateFrame):
            frame = frame.replicate_without_data(**frame_kwargs)
        elif callable(frame):
            frame = frame(**frame_kwargs)
        else:
            raise TypeError(f"Got invalid frame of type {type(frame)}: {frame}")

        return frame

    def transform_to(self, frame: CoordFrameType, /) -> "Coordinate":
        """Transform celestial coordinate between any frames.

        No time conversion is performed, so this operation on `Coordinate` object with
        no `time` attached can fail, depending on the target frame.

        Parameters
        ----------
        frame
            The frame to convert to.

        Returns
        -------
        The converted coordinates.

        """
        frame = to_astropy_type.frame(frame)
        frame = self._normalize_frame(frame)

        if (self.time is None) and (
            (getattr(self.frame, "obstime", ...) is None)
            or (getattr(frame, "obstime", ...) is None)
        ):
            raise ValueError(
                "Cannot transform coordinate without time attached to it. "
                "Please attach a time to the coordinate, or use `transform_to_time` "
                "to transform to a frame with a specific time."
            )

        # Interpolate astrometric parameters, for better performance.
        # https://docs.astropy.org/en/stable/coordinates/index.html#improving-performance-for-arrays-of-obstime
        with erfa_astrom.set(ErfaAstromInterpolator(300 * u.s)):
            return self.__class__.from_skycoord(self.skycoord.transform_to(frame))

    def cartesian_offset_by(self, offset: "CoordinateDelta", /) -> "Coordinate":
        """Calculate coordinates at the given offset from this coordinate.

        Parameters
        ----------
        offset
            Offset from this coordinate.

        Returns
        -------
        Coordinates at the given offset.

        Examples
        --------
        >>> coord = calc.coord.from_name("Sun", time.time())
        >>> coord.cartesian_offset_by(
        ...     neclib.coordinates.CoordinateDelta(1 * u.deg, 1 * u.deg))
        Coordinate(
            lon=<Longitude 6.20805658 rad>,
            lat=<Latitude -0.03254802 rad>,
            frame=<GCRS Frame (
                obstime=1678968349.001554,
                obsgeoloc=(-3548179.23395423, 3753814.11380175, 3731512.91152004) m,
                obsgeovel=(-273.72298624, -259.34513153, 0.62044525) m / s)>,
            time=<Time object: scale='utc' format='unix' value=1678968349.001554>)

        """
        if not isinstance(offset, CoordinateDelta):
            raise TypeError(
                f"Expected `CoordinateDelta` object, but got {type(offset).__name__}"
            )
        coord_in_offset_frame = self.transform_to(offset.frame)
        lon: u.Quantity = coord_in_offset_frame.lon + offset.d_lon  # type: ignore
        lat: u.Quantity = coord_in_offset_frame.lat + offset.d_lat  # type: ignore
        return self.__class__(
            lon=lon, lat=lat, distance=self.distance, frame=offset.frame, time=self.time
        )

    @classmethod
    def from_skycoord(cls, coord: SkyCoord, /):
        return cls(
            lon=coord.spherical.lon,  # type: ignore
            lat=coord.spherical.lat,  # type: ignore
            distance=getattr(coord.data, "distance", None),
            frame=coord.frame.replicate_without_data(),
            time=getattr(coord, "obstime", None),
        )

    def replicate(self, **kwargs: Any):
        _fields = {
            f.name: kwargs.get(f.name, getattr(self, f.name)) for f in fields(self)
        }
        return self.__class__(**_fields)

    def to_apparent_altaz(self) -> "ApparentAltAzCoordinate":
        """Convert celestial coordinate in any frame to telescope frame.

        This method converts the given ``coord`` to AltAz frame, taking into account
        the pointing error correction and the sidereal motion.

        Returns
        -------
        The converted coordinates, in AltAz frame.

        Notes
        -----
        If the coordinate object is already in the AltAz frame, this method simply
        returns the broadcasted version of it. This is because coordinates in AltAz
        frame should have already accounted for the sidereal motion. Therefore, no
        additional tracking calculations are necessary, only pointing error correction.

        """
        if self.time is None:
            raise ValueError("time is not given.")

        if self.frame.name == "altaz":  # type: ignore
            altaz = self.broadcasted
        else:
            altaz = self.transform_to("altaz")

        az, alt = self._calc.pointing_err.refracted_to_apparent(altaz.lon, altaz.lat)
        return ApparentAltAzCoordinate(az=az, alt=alt, time=self.time)

    @property
    def broadcasted(self) -> "Coordinate":
        if self.time is None:
            return self

        if self.lon.shape != self.lat.shape:
            raise ValueError(
                "`lon` and `lat` must have the same shape, but are "
                f"{self.lon.shape} and {self.lat.shape}."
            )
        elif self.lon.shape == self.time.shape:
            return self

        if self.lon.isscalar or (self.lon.size == 1):
            lon: u.Quantity = (
                np.broadcast_to(self.lon, self.time.shape) << self.lon.unit
            )
            lat: u.Quantity = (
                np.broadcast_to(self.lat, self.time.shape) << self.lat.unit
            )
            distance: Optional[u.Quantity] = (
                None
                if self.distance is None
                else np.broadcast_to(self.distance, self.time.shape)  # type: ignore
            )
            time = self.time
        elif self.time.isscalar or (self.time.size == 1):
            lon = self.lon
            lat = self.lat
            distance = self.distance
            time: Time = np.broadcast_to(self.time, self.lon.shape)  # type: ignore
        else:
            raise ValueError(
                "Either `lon` or `lat` must be a scalar, or they must have the same "
                "shape as `time`."
            )
        return self.__class__(
            lon=lon, lat=lat, distance=distance, frame=self.frame, time=time
        )

    @property
    def size(self) -> int:
        return self.broadcasted.lon.size

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.broadcasted.lon.shape


class NameCoordinate(Coordinate):
    def __init__(self, name: str, time: Optional[Time] = None, /):
        self.name = name
        self.time = time

    def realize(
        self, time: Optional[Union[int, float, Array[Union[int, float]], Time]] = None
    ) -> Coordinate:
        """Get the position of a celestial body from its name.

        Parameters
        ----------
        name
            Name of the celestial body.
        obstime
            Time of observation.

        Returns
        -------
        Position (SkyCoord) of the celestial body.

        Examples
        --------
        >>> calc.coord.from_name("Sun", time.time())
        Coordinate(
            lon=<Longitude 6.20805658 rad>,
            lat=<Latitude -0.03254802 rad>,
            frame=<GCRS Frame (
                obstime=1678968349.001554,
                obsgeoloc=(-3548179.23395423, 3753814.11380175, 3731512.91152004) m,
                obsgeovel=(-273.72298624, -259.34513153, 0.62044525) m / s)>,
            time=<Time object: scale='utc' format='unix' value=1678968349.001554>)
        >>> calc.coord.from_name("Orion KL", [time.time(), time.time() + 1])
        Coordinate(
            lon=<Longitude 83.809 deg>,
            lat=<Latitude -5.372639 deg>,
            frame=<ICRS Frame>,
            time=<Time object:
                scale='utc' format='unix' value=[1.67896831e+09 1.67896831e+09]>)

        """
        time = to_astropy_type.time(time)
        if time is None:
            time = self.time

        try:
            coord = get_body(self.name, time, location=self._calc.location)
        except KeyError:
            coord = SkyCoord.from_name(self.name)
        ret = self._calc.coordinate.from_skycoord(coord)
        if (ret.time is None) and (time is not None):
            ret = ret.replicate(time=time)
        return ret


@dataclass
class ApparentAltAzCoordinate:
    az: Union[DimensionLess, u.Quantity]
    alt: Union[DimensionLess, u.Quantity]
    time: Union[DimensionLess, Time]
    unit: Optional[UnitType] = None

    def __post_init__(self) -> None:
        self.az = get_quantity(self.az, unit=self.unit)
        self.alt = get_quantity(self.alt, unit=self.unit)
        self.time = to_astropy_type.time(self.time)

    @property
    def broadcasted(self) -> "ApparentAltAzCoordinate":
        if self.time is None:
            return self

        if self.az.shape != self.alt.shape:
            raise ValueError(
                "`az` and `alt` must have the same shape, but are "
                f"{self.az.shape} and {self.alt.shape}."
            )
        elif self.az.shape == self.time.shape:
            return self

        if self.az.isscalar:
            lon: u.Quantity = np.broadcast_to(self.az, self.time.shape) << self.az.unit
            lat: u.Quantity = (
                np.broadcast_to(self.alt, self.time.shape) << self.alt.unit
            )
            time = self.time
        elif self.time.isscalar:
            lon = self.az
            lat = self.alt
            time: Time = np.broadcast_to(self.time, self.az.shape)  # type: ignore
        else:
            raise ValueError(
                "Either `lon` or `lat` must be a scalar, or they must have the same "
                "shape as `time`."
            )
        return self.__class__(az=lon, alt=lat, time=time)

    @property
    def size(self) -> int:
        return self.broadcasted.az.size

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.broadcasted.alt.shape


@dataclass
class CoordinateDelta:
    d_lon: Union[DimensionLess, u.Quantity]
    d_lat: Union[DimensionLess, u.Quantity]
    frame: Union[Type[BaseCoordinateFrame], BaseCoordinateFrame, str]
    unit: Optional[UnitType] = None

    def __post_init__(self) -> None:
        """Create a coordinate delta from builtin type values."""
        self.frame = to_astropy_type.frame(self.frame)
        self.d_lon = get_quantity(self.d_lon, unit=self.unit)
        self.d_lat = get_quantity(self.d_lat, unit=self.unit)

    @property
    def broadcasted(self) -> "CoordinateDelta":
        if self.d_lon.shape != self.d_lat.shape:
            raise ValueError(
                "`d_lon` and `d_lat` must have the same shape, but are "
                f"{self.d_lon.shape} and {self.d_lat.shape}."
            )
        return self.__class__(d_lon=self.d_lon, d_lat=self.d_lat, frame=self.frame)

    @property
    def size(self) -> int:
        return self.broadcasted.d_lon.size

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.broadcasted.d_lon.shape


@dataclass
class CoordCalculator:
    """Collection of basic methods for celestial coordinate calculation.

    Parameters
    ----------
    location
        Location of the telescope.
    pointing_param_path
        Path to the pointing error correction parameters. If None, pointing error
        correction won't be performed.

    Examples
    --------
    >>> calc = neclib.coordinates.CoordCalculator(config.location)

    """

    location: EarthLocation = config.location  # type: ignore
    pointing_err_file: Optional[Union[os.PathLike, str]] = None

    obswl: ClassVar[QuantityValidator] = QuantityValidator(unit="mm")
    obsfreq: ClassVar[QuantityValidator] = QuantityValidator(
        config.observation_frequency, unit="GHz"  # type: ignore
    )
    relative_humidity: ClassVar[QuantityValidator] = QuantityValidator(unit="")
    pressure: ClassVar[QuantityValidator] = QuantityValidator(unit="hPa")
    temperature: ClassVar[QuantityValidator] = QuantityValidator(unit="K")

    command_group_duration_sec = 1

    @property
    def command_freq(self) -> Union[int, float]:
        return config.antenna_command_frequency  # type: ignore

    @property
    def command_offset_sec(self) -> Union[int, float]:
        return config.antenna_command_offset_sec  # type: ignore

    @property
    def pointing_err(self) -> PointingError:
        if not hasattr(self, "_pointing_err"):
            if self.pointing_err_file is None:
                logger.warning("Pointing error correction is disabled. ")
                self._pointing_err = PointingError.get_dummy()
            else:
                self._pointing_err = PointingError.from_file(self.pointing_err_file)
        return self._pointing_err

    @property
    def altaz_kwargs(self) -> Dict[str, Any]:
        """Return keyword arguments for AltAz frame, except for ``obstime``."""
        # Check if diffraction correction is enabled.
        _diffraction_params = ("pressure", "temperature", "relative_humidity", "obswl")
        diffraction_params = {x: getattr(self, x) for x in _diffraction_params}
        not_set = [k for k, v in diffraction_params.items() if v != v]
        if len(not_set) > 0:
            logger.warning(
                f"Diffraction correction is disabled. {not_set} are not given."
            )

        # Check if obswl and obsfreq are consistent.
        obswl = None
        if self.obswl == self.obswl:  # Check if self.obswl is NaN or not.
            obswl = self.obswl
        if self.obsfreq == self.obsfreq:  # Check if self.obsfreq is NaN or not.
            _obswl = self.obsfreq.to(u.mm, equivalencies=u.spectral())
            if (obswl is not None) and (obswl != _obswl):
                raise ValueError(
                    f"obswl={obswl} and obsfreq={self.obsfreq} are inconsistent."
                )
            obswl = _obswl

        return dict(
            location=self.location,
            temperature=self.temperature.to(u.deg_C, equivalencies=u.temperature()),
            pressure=self.pressure,
            relative_humidity=self.relative_humidity,
            obswl=obswl,
        )

    @property
    def coordinate(self) -> Type[Coordinate]:
        Coordinate._calc = self
        return Coordinate

    @property
    def name_coordinate(self) -> Type[NameCoordinate]:
        NameCoordinate._calc = self
        return NameCoordinate

    coordinate_delta = CoordinateDelta



from abc import ABC, abstractmethod
from collections import UserDict
from typing import (
    Any,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    final,
    overload,
)

from .. import config, utils
from ..core import Parameters
from ..core.configuration import Configuration
from ..core.normalization import partial


def get_device_configuration():
    return {k[:-2]: v for k, v in config.items() if k.endswith("._")}


class DeviceBase(ABC):

    Model: ClassVar[str]
    Manufacturer: ClassVar[str]
    Identifier: ClassVar[Optional[str]] = None
    is_simulator: ClassVar[bool] = False

    Config: ClassVar[Union[Configuration, None]] = None

    _instances: ClassVar[Dict[Any, "DeviceBase"]]
    _initialized: ClassVar[bool] = False
    _implementations: List[Type["DeviceBase"]] = []
    _is_autogenerated: ClassVar[bool] = False
    _kind: ClassVar[Type["DeviceBase"]]

    def __init_subclass__(cls) -> None:
        if hasattr(cls, "Model") and (not cls._is_autogenerated):
            # `cls.__subclasses__()` returns its immediate subclasses only, and they
            # should be device kind definition, not controller implementations for each
            # model. This (`cls._implementations`) is similar to that, but this includes
            # grandchild classes, and not includes abstract classes (no `cls.Model` is
            # set).
            cls._implementations.append(cls)
            if not hasattr(cls, "_instances"):
                setattr(cls, "_instances", {})
        if cls.__base__ is DeviceBase:
            cls._kind = cls

    def __new__(cls) -> "DeviceBase":
        if (cls is DeviceBase) or (not hasattr(cls, "Model")):
            # Ensure no instantiation of ABC, to avoid unintended class variable change.
            raise TypeError

        # Singleton check.
        identity = cls._identity(cls.Config, cls.Identifier)
        if identity not in cls._instances:
            inst = super().__new__(cls)
            cls._instances[identity] = inst

        # Initialization status check.
        if cls._initialized:
            cls.__init__ = lambda *args, **kwargs: None
        cls._initialized = True

        return cls._instances[identity]

    @classmethod
    def get_simulator_class(cls) -> Type["DeviceBase"]:
        if cls.is_simulator:
            return cls
        same_kind = cls._kind.__subclasses__()
        simulator = [c for c in same_kind if c.is_simulator]
        if len(simulator) == 0:
            raise NotImplementedError(
                f"No simulator implementation for {cls.__name__} "
                f"({cls._kind.__name__}) is found."
            )
        return simulator[0]

    @staticmethod
    def _normalize(key: str, /) -> str:
        return key.replace("_", "").lower()

    @classmethod
    def _find_config(
        cls, name: str, identifier: Optional[str] = None
    ) -> Optional[Configuration]:
        this_device_config = config[name]
        model = cls._normalize(this_device_config._)  # type: ignore
        identity = None
        if identifier is not None:
            identity = getattr(this_device_config, identifier, None)

        device_configuration = get_device_configuration()
        model_filtered = [
            k for k, v in device_configuration.items() if cls._normalize(v) == model
        ]
        _cfgs: List[Configuration] = [config[k] for k in model_filtered]  # type: ignore
        cfg = None
        for _cfg in _cfgs:
            if (identifier is None) or (getattr(_cfg, identifier, None) == identity):
                cfg += _cfg
        return cfg

    @staticmethod
    def _identity(
        cfg: Union[Configuration, None], identifier: Optional[str] = None
    ) -> Any:
        if identifier is None:
            return None
        identity = getattr(cfg, identifier, None)
        if isinstance(identity, Parameters):
            return None
        return identity

    @overload
    @classmethod
    def bind(cls, name: str, model: str) -> Type["DeviceBase"]:
        ...

    @overload
    @classmethod
    def bind(cls, name: str, model: Dict[str, str]) -> "Devices[str, Type[DeviceBase]]":
        ...

    @final
    @classmethod
    def bind(
        cls, name: str, model: Union[str, Dict[str, str]]
    ) -> Union[Type["DeviceBase"], "Devices[str, Type[DeviceBase]]"]:
        impl = {cls._normalize(_impl.Model): _impl for _impl in cls._implementations}
        if isinstance(model, dict):
            bound_devices = {
                key: cls.bind(f"{name}.{key}", _model) for key, _model in model.items()
            }
            return Devices(bound_devices)
        else:
            model_impl = impl[cls._normalize(model)]
            if config.simulator:
                try:
                    simulator = model_impl.get_simulator_class()
                    simulator.Identifier = model_impl.Identifier
                    model_impl = simulator
                except NotImplementedError:
                    pass
            Name = ".".join([utils.toCamelCase(n) for n in name.split(".")])
            cfg = cls._find_config(name, model_impl.Identifier)
            identity = cls._identity(cfg, model_impl.Identifier)
            if identity in model_impl._instances:
                return model_impl._instances[identity].__class__
            return type(
                Name,
                (model_impl,),
                dict(Config=cfg, __module__=cls.__module__, _is_autogenerated=True),
            )

    def __repr__(self) -> str:
        model = f"model={self.Model}"
        manufacturer = f"manufacturer={self.Manufacturer}"
        identity = "<no device identity defined>"
        if self.Identifier:
            _id = getattr(self.Config, self.Identifier, None)
            identity = f"{self.Identifier}={_id!r}"
        metadata = ", ".join([model, manufacturer, identity])
        return f"{self.__class__.__name__}({metadata})"

    __str__ = __repr__

    @abstractmethod
    def finalize(self) -> None:
        ...


T_key = TypeVar("T_key")
T_value = TypeVar("T_value")


class Devices(UserDict, Generic[T_key, T_value]):
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({super().__repr__()})"

    __str__ = __repr__

    def __init__(
        self,
        anonymous_device: Optional[
            Union[
                DeviceBase,
                Type[DeviceBase],
                Dict[str, DeviceBase],
                Dict[str, Type[DeviceBase]],
            ]
        ] = None,
        /,
        **named_devices: Union[DeviceBase, Type[DeviceBase]],
    ) -> None:
        super().__init__()

        if (anonymous_device is not None) and (len(named_devices) > 0):
            raise ValueError("Cannot specify both anonymous and named devices")

        if anonymous_device is None:
            self.update(named_devices)
        elif isinstance(anonymous_device, dict):
            # Keep compatibility with `dict` constructor's signature;
            # `__init__(__map, /, **kwargs) -> None`
            self.update(anonymous_device)
        else:
            self.update({None: anonymous_device})

    @property
    def _is_single_anonymous_device(self) -> bool:
        return (len(self) == 1) and (list(self.keys())[0] is None)

    def _parse_id_query(
        self, id: Optional[str] = None
    ) -> Tuple[Union[None, str], Union[None, str]]:
        if id is None:
            return (None, None)
        if self._is_single_anonymous_device:
            # Anonymous device doesn't have device-ID by definition
            return (None, id)
        if id.find(".") != -1:
            # Dot-separated value is parsed to `<device-ID>.<channel-ID>`
            return tuple(id.rsplit(".", 1))
        # Device-ID takes precedence when given `id` doesn't match certain pattern
        return (id, None)

    @overload
    def __call__(self, *args, id: str, **kwargs) -> DeviceBase:
        ...

    @overload
    def __call__(self, *args, **kwargs) -> "Devices":
        ...

    def __call__(self, *args, **kwargs) -> Union[DeviceBase, "Devices"]:
        """Emulate initialization of attached device controllers."""
        device_id, channel_id = self._parse_id_query(kwargs.get("id", None))
        if (device_id is None) and (channel_id is None):
            return Devices({k: v(*args, **kwargs) for k, v in self.items()})
        kwargs["id"] = channel_id
        return self[device_id](*args, **kwargs)

    def __getattr__(self, key: str, /) -> Any:
        """Emulate attribute access to attached device controllers."""
        targets = {k: getattr(v, key) for k, v in self.items()}
        if all(callable(t) for t in targets.values()):

            def func(*args, **kwargs) -> Union[Any, Dict[Union[str, None], Any]]:
                device_id, channel_id = self._parse_id_query(kwargs.get("id", None))
                if (device_id is None) and (channel_id is None):
                    if self._is_single_anonymous_device:
                        return targets[None](*args, **kwargs)
                    return {k: t(*args, **kwargs) for k, t in targets.items()}
                kwargs["id"] = channel_id
                return targets[device_id](*args, **kwargs)

            return func

        elif self._is_single_anonymous_device:
            return targets[None]
        else:
            return targets

    def __getitem__(self, key: Union[str, None], /) -> Any:
        try:
            return super().__getitem__(key)
        except KeyError:
            pass

        if isinstance(key, str):
            item_normalized = {
                k.replace("_", "").lower() if isinstance(k, str) else k: v
                for k, v in self.items()
            }
            _key = key.replace("_", "").lower()
            if _key in item_normalized:
                return item_normalized[_key]

            device_id, channel_id = self._parse_id_query(_key)
            if device_id in item_normalized:
                item = item_normalized[device_id]
                for attr_name in dir(item):
                    attr = getattr(item, attr_name, None)
                    if attr is None:
                        continue
                    try:
                        modified = partial(attr, kwargs={"id": channel_id})
                        print(attr_name, modified)
                        object.__setattr__(item, attr_name, modified)
                    except TypeError as e:
                        print(attr_name, e)
                        continue
                return item

            raise KeyError(channel_id)

        raise KeyError(key)



from __future__ import annotations

from io import StringIO
from typing import TYPE_CHECKING

from ..inform import get_logger
from ..types import CoordFrameType

if TYPE_CHECKING:
    # Circular import
    from ...coordinates.observations.observation_spec_base import ObservationSpec

__all__ = ["html_repr_of_observation_spec"]

logger = get_logger(__name__)


def html_repr_of_observation_spec(
    observation_spec: ObservationSpec, /, *, frame: CoordFrameType = "fk5"
) -> str:
    """Return HTML representation of the observation specification."""

    waypoint_repr = []

    _svg = StringIO()
    observation_spec.fig.savefig(_svg, format="svg")
    _svg.seek(0)
    svg = _svg.read()

    counter = 0
    for _, _wp in observation_spec.coords.groupby(level=0):
        if counter > 30:  # Limit the number of waypoints to be displayed.
            waypoint_repr.append('<tr><td colspan="3">...</td></tr>')
            break

        for mode, wp in _wp.groupby("mode", sort=False):
            counter += 1

            lon, lat = wp["lon"], wp["lat"]
            try:
                _coord_str = [
                    f"({_lon:9.4f}, {_lat:9.4f})" for _lon, _lat in zip(lon, lat)
                ]
                coord_str = " &#x2192; ".join(_coord_str)
            except TypeError:
                coord_str = f"({lon:9.4f}, {lat:9.4f})"
            others = []
            for info in ["speed", "scan_frame", "integration", "id"]:
                attr, *_ = wp[info].unique()
                if attr is not None:
                    others.append(f"{info}={attr}")
            waypoint_repr.append(
                f"""
                <tr>
                    <td>{mode.name}</td>
                    <td><code>{coord_str}</code></td>
                    <td>{", ".join(others)}</td>
                </tr>
                """
            )

    return f"""
    <details><summary>{observation_spec.coords.index[-1]} waypoints</summary>
        <table>
            <thead>
                <tr>
                    <th>Mode</th>
                    <th>{frame.upper()} Coordinate</th>
                    <th>Other Parameters</th>
                </tr>
            </thead>
            <tbody>
                {''.join(waypoint_repr)}
            </tbody>
        </table>
        <span>The time interval for OFF and HOT observations are not considered.</span>
    </details>
    <details open><summary>Graph</summary>
        {svg}
    </details>
    """



import math
import time
from dataclasses import dataclass
from itertools import count
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import astropy.units as u

from ..core.types import CoordFrameType, DimensionLess, UnitType
from . import paths
from .convert import CoordCalculator


@dataclass
class ApparentAltAzCoordinate:
    az: u.Quantity
    """Azimuth angle."""
    el: u.Quantity
    """Elevation angle."""
    time: List[float]
    """Time for each coordinate."""
    context: paths.ControlContext
    """Metadata of the control section this coordinate object is a part of."""


T = TypeVar("T", bound=Union[DimensionLess, u.Quantity])
CoordinateGenerator = Generator[ApparentAltAzCoordinate, Literal[True], None]


class PathFinder(CoordCalculator):
    @overload
    def from_function(
        self,
        lon: Callable[[paths.Index], T],
        lat: Callable[[paths.Index], T],
        frame: CoordFrameType,
        /,
        *,
        unit: Optional[UnitType] = None,
        n_cmd: Union[int, float],
        context: paths.ControlContext,
    ) -> CoordinateGenerator:
        ...

    @overload
    def from_function(
        self,
        lon_lat: Callable[[paths.Index], Tuple[T, T]],
        frame: CoordFrameType,
        /,
        *,
        unit: Optional[UnitType] = None,
        n_cmd: Union[int, float],
        context: paths.ControlContext,
    ) -> CoordinateGenerator:
        ...

    def from_function(
        self,
        *coord: Union[
            Callable[[paths.Index], T],
            Callable[[paths.Index], Tuple[T, T]],
            CoordFrameType,
        ],
        unit: Optional[UnitType] = None,
        n_cmd: Union[int, float],
        context: paths.ControlContext,
    ) -> CoordinateGenerator:
        """Generate coordinate commands from arbitrary function."""
        if len(coord) == 3:
            lon_func, lat_func, frame = coord

            def lon_lat_func(idx: paths.Index) -> Tuple[T, T]:
                return lon_func(idx), lat_func(idx)  # type: ignore

        elif len(coord) == 2:
            lon_lat_func, frame = coord  # type: ignore
        else:
            raise TypeError(
                "Invalid number of positional arguments: expected 2 ("
                "func(idx)->(lon, lat) and coordinate_frame) or 3 (for func1(idx)->lon,"
                f" func2(idx)->lat and coordinate_frame), but got {len(coord)}"
            )

        unit_n_cmd = int(self.command_group_duration_sec * self.command_freq)
        if context.start is None:
            context.start = time.time() + self.command_offset_sec
        if context.stop is None:
            context.stop = context.start + n_cmd / self.command_freq

        for seq in range(math.ceil(n_cmd / unit_n_cmd)):
            start_idx = seq * unit_n_cmd
            _idx = [start_idx + i for i in range(unit_n_cmd) if start_idx + i <= n_cmd]
            _t = [context.start + i / self.command_freq for i in _idx]
            idx = paths.Index(time=_t, index=_idx)

            lon_for_this_seq, lat_for_this_seq = lon_lat_func(idx)
            _coord = self.coordinate(
                lon=lon_for_this_seq,
                lat=lat_for_this_seq,
                frame=frame,  # type: ignore
                unit=unit,
                time=idx.time,
            )
            altaz = _coord.to_apparent_altaz()
            sent = yield ApparentAltAzCoordinate(
                az=altaz.az,  # type: ignore
                el=altaz.alt,  # type: ignore
                time=_t,
                context=context,
            )
            if (sent is not None) and context.waypoint:
                context.stop = idx.time[-1]
                break

    def sequential(
        self,
        *section_args: Tuple[Sequence[Any], Dict[str, Any]],
        repeat: Union[int, Sequence[int]] = 1,
    ) -> CoordinateGenerator:
        if isinstance(repeat, int):
            counter = [range(repeat) if repeat > 0 else count()] * len(section_args)
        else:
            counter = [range(n) if n > 0 else count() for n in repeat]

        last_stop = None
        to_break = False

        ctx = paths.ControlContext()

        for c, (args, kwargs) in zip(counter, section_args):
            for _ in c:
                context: paths.ControlContext = kwargs["context"]
                ctx.update(context)

                # Independent path calculators may not know when the computed command
                # will be sent, especially when the path follows another path. They just
                # know how long it takes to complete the commands they computed. So the
                # time consistency is computed here.
                if context.duration is not None:
                    context.start = last_stop or time.time() + self.command_offset_sec
                    context.stop = context.start + context.duration
                last_stop = context.stop

                section = self.from_function(*args, **kwargs)
                for coord in section:
                    sent = yield coord
                    if (sent is not None) and coord.context.waypoint:
                        to_break = True
                        context.stop = last_stop = coord.time[-1]

                if to_break:
                    to_break = False
                    break

    def linear(
        self,
        *target: Union[DimensionLess, u.Quantity, str, CoordFrameType],
        unit: Optional[UnitType] = None,
        start: Tuple[T, T],
        stop: Tuple[T, T],
        scan_frame: CoordFrameType,
        speed: T,
        margin: Optional[T] = None,
        offset: Optional[Tuple[T, T, CoordFrameType]] = None,
    ) -> CoordinateGenerator:
        args = (self, *target)
        kwargs = dict(
            unit=unit,
            start=start,
            stop=stop,
            scan_frame=scan_frame,
            speed=speed,
            margin=margin,
            offset=offset,
        )
        path1 = paths.Standby(*args, **kwargs)  # type: ignore
        path2 = paths.Accelerate(*args, **kwargs)  # type: ignore
        path3 = paths.Linear(*args, **kwargs)  # type: ignore
        arguments1 = path1.arguments
        arguments2 = path2.arguments
        arguments3 = path3.arguments

        yield from self.sequential(
            arguments1, arguments2, arguments3, repeat=[-1, 1, 1]
        )

    def track(
        self,
        *target: Union[DimensionLess, u.Quantity, str, CoordFrameType],
        unit: Optional[UnitType] = None,
        offset: Optional[Tuple[T, T, CoordFrameType]] = None,
        **ctx_kw: Any,
    ) -> CoordinateGenerator:
        path = paths.Track(self, *target, unit=unit, offset=offset, **ctx_kw)
        arguments = path.arguments
        yield from self.sequential(arguments, repeat=-1)


class CoordinateGeneratorManager:
    def __init__(self, generator: Optional[CoordinateGenerator] = None) -> None:
        self._generator = generator
        self._send_value = None

    def will_send(self, value: Any) -> None:
        self._send_value = value

    def __iter__(self) -> Iterable[Any]:
        return self  # type: ignore

    def __next__(self) -> Any:
        if self._generator is None:
            self._send_value = None
            raise StopIteration("No generator attached")
        if self._send_value is None:
            return next(self._generator)
        try:
            ret = self._generator.send(self._send_value)
            self._send_value = None
            return ret
        except TypeError:
            # Keep send value once for just-started generator
            return next(self._generator)

    def attach(self, generator: CoordinateGenerator) -> None:
        self.clear()
        self._generator = generator

    def clear(self) -> None:
        if self._generator is not None:
            try:
                self._generator.close()
            except Exception:
                pass
        self._generator = None

    def get(self) -> Optional[CoordinateGenerator]:
        return self._generator



import time
from typing import Any, Callable, Generic, Optional, Type, TypeVar

from ..formatting import html_repr_of_status

T = TypeVar("T", bound=Type)


class StatusManager(Generic[T]):
    """Manage multiple contexts.

    Parameters
    ----------
    ctx_type
        A dataclass type which keeps single status/context. The arguments of its
        constructor should have default values, otherwise ``get(out_of_range)`` will
        fail.
    start
        Name of the attribute which stores the start index of the context. The index
        would typically be the time.
    stop
        Name of the attribute which stores the stop index of the context.
    idx_getter
        A function which returns the current index. This function is called with no
        arguments.
    keep
        Maximum index difference from current one to keep the context. Contexts with
        index difference larger than this value will be removed.

    Examples
    --------
    >>> @dataclass
    ... class Context:
    ...     start: Optional[float] = None
    ...     stop: Optional[float] = None
    ...     value: int = 0
    >>> manager = StatusManager(Context)
    >>> manager.get(time.time())
    Context(start=None, stop=None, value=0)  # default value
    >>> manager.set(start=time.time(), stop=time.time() + 10, value=100)
    >>> manager.get(time.time())
    Context(start=..., stop=..., value=100)

    """

    def __init__(
        self,
        ctx_type: Type[T],
        /,
        start: str = "start",
        stop: str = "stop",
        idx_getter: Callable[[], float] = time.time,
        keep: float = 60,
    ) -> None:
        self.ctx = []
        self.start = start
        self.stop = stop
        self.ctx_type = ctx_type
        self.idx_getter = idx_getter
        self.keep = keep

    def _sort(self) -> None:
        self.ctx.sort(key=lambda x: getattr(x, self.start))

        # Remove stale contexts.
        current_idx = self.idx_getter()
        threshold = current_idx - self.keep
        outdated = [ctx for ctx in self.ctx if getattr(ctx, self.stop) < threshold]
        [self.ctx.remove(ctx) for ctx in outdated]

    def get(self, idx: float, /) -> T:
        """Find the context that contains the specified index value.

        Parameters
        ----------
        idx : float
            The index value to search for.

        Returns
        -------
        The context that contains the specified index value.

        """
        not_future = [ctx for ctx in self.ctx if getattr(ctx, self.start) <= idx]
        for ctx in reversed(not_future):
            # This search is done from the end of the context list so that the newer
            # context takes precedence. Caution this is not the context setting order.
            if (getattr(ctx, self.stop) is None) or (idx < getattr(ctx, self.stop)):
                return ctx

        # If no context is found, return the default values.
        return self.ctx_type()

    def set(self, _obj: Optional[T] = None, /, **kwargs: Any) -> None:
        """Set the context.

        Parameters
        ----------
        kwargs
            Keyword arguments to be passed to the constructor of the context type.

        """
        if _obj is not None:
            self.ctx.append(_obj)
        else:
            ctx = self.ctx_type(**kwargs)
            self.ctx.append(ctx)
        self._sort()

    def _repr_html_(self) -> str:
        return html_repr_of_status(self)



import functools
from inspect import Parameter, Signature, signature
from types import FunctionType, MethodType
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

from ..exceptions import NotInitializedError


class partial:
    def __init__(
        self,
        func: Any = None,
        /,
        *,
        args: Tuple[Any, ...] = (),
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        if isinstance(func, partial):
            func = func.func

        if (func is not None) and (not isinstance(func, (MethodType, FunctionType))):
            raise TypeError(
                f"Only function-like type supported, given {func} ({type(func)})"
            )
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(
        self, *args: Any, **kwargs: Any
    ) -> Union[Callable[..., Any], "partial"]:
        if (self.func is None) and (len(args) > 0):
            self.__init__(args[0], args=self.args, kwargs=self.kwargs)
            return self
        else:
            return self.create_wrapper(shift=0)(*args, **kwargs)

    def __get__(
        self, obj: object, objtype: Optional[Type[object]] = None
    ) -> Callable[..., Any]:
        return self.create_wrapper(shift=0 if obj is None else 1)

    def empty_parameters(
        self, args: Tuple[Any, ...], kwargs: Dict[str, Any], shift: int = 0
    ) -> List[str]:
        sig = self.signature
        empty_args = [
            list(sig.parameters.items())[i][0]
            for i, v in enumerate(args)
            if v == Parameter.empty
        ]
        empty_kwargs = [k for k, v in kwargs.items() if v == Parameter.empty]
        empty = empty_args + empty_kwargs
        return empty[min(len(empty), shift) :]

    def create_wrapper(self, shift: int = 0) -> Callable[..., Any]:
        @functools.wraps(self.func)
        def wrapper(*_args, **_kwargs):
            args_1 = self.parse_arguments(self.args, self.kwargs, True, shift)
            args_2 = self.parse_arguments(_args, _kwargs, False, shift)
            args, kwargs = self.merge_arguments(args_1, args_2)
            empty = self.empty_parameters(args, kwargs, shift=shift)
            if len(empty) > 0:
                raise TypeError(f"Values for argument {empty} not provided")
            return self.func(*args, **kwargs)

        return wrapper

    @property
    def signature(self) -> Signature:
        if self.func is None:
            raise NotInitializedError("Not attached to function-like object.")
        return signature(self.func)

    def parse_arguments(
        self,
        args: Tuple[Any, ...] = (),
        kwargs: Optional[Dict[str, Any]] = None,
        fallback: bool = False,
        shift: int = 0,
    ) -> Dict[Union[str, int], Any]:
        kwargs = kwargs or {}
        arguments = {}
        for i, param in enumerate(self.signature.parameters.values()):
            if param.kind == Parameter.POSITIONAL_ONLY:
                arguments[i] = param.default if fallback else Parameter.empty
            else:
                arguments[param.name] = param.default if fallback else Parameter.empty

        not_kw_only = filter(
            lambda x: x.kind != Parameter.KEYWORD_ONLY,
            self.signature.parameters.values(),
        )
        possibly_positional_count = len(list(not_kw_only))

        keys = tuple(arguments.keys())
        for i, arg in enumerate(args):
            if i < possibly_positional_count:
                k = keys[i + shift]
                arguments[k] = arg
        for k, kwarg in kwargs.items():
            if k in arguments:
                arguments[k] = kwarg

        return arguments

    def merge_arguments(
        self, *args: Dict[Union[str, int], Any]
    ) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        """Merge arguments candidates.

        Parameters
        ----------
        args
            The (args, kwargs) pairs to be merged. The pair appears later takes
            precedence, so the default value would be given as first argument.

        """
        arguments = self.parse_arguments(fallback=True)
        for _args in args:
            not_empty = dict(
                filter(lambda item: item[1] != Parameter.empty, _args.items())
            )
            arguments.update(not_empty)
        return tuple(v for k, v in arguments.items() if isinstance(k, int)), {
            k: v for k, v in arguments.items() if isinstance(k, str)
        }



from typing import Tuple

import astropy.units as u
import numpy as np

from .pointing_error import PointingError


class OMU1P85M(PointingError):
    r"""Pointing model used in OMU1P85M.

    .. math::

        \Delta x =& a_1 \sin ( El ) \\
        &+ a_2 \\
        &+ a_3 \cos ( El ) \\
        &+ b_1 \sin ( Az ) \sin ( El ) \\
        &- b_2 \cos ( Az ) \sin ( El ) \\
        &+ c_1 * \sin ( Az - El ) \\
        &+ c_2 * \cos ( Az - El ) \\
        &+ d_1 \\
        &+ e_1 \cos ( El ) \\
        &- e_2 \sin ( El ) \\
        \Delta Az =& - \Delta x / \cos ( El ) \\
        \Delta y =& b_1 \cos ( Az ) \\
        &+ b_2 \sin ( Az ) \\
        &+ b_3 \\
        &+ g_1 El \\
        &+ c_1 \cos ( Az - El ) \\
        &- c_2 \sin ( Az - El ) \\
        &+ d_2 \\
        &+ e_1 \sin ( El ) \\
        &+ e_2 \cos ( El ) \\
        \Delta El =& - \Delta y

    Parameters
    ----------
    a1
        Skew angle (lack of orthogonality) between azimuth and elevation axes.
    a2
        X collimation error; non-orthogonality between elevation and optical axes.
    a3
        Azimuth (not X) offset of encoder reading.
    b1
        Tilt angle of azimuth axis, in North-South direction. North-positive.
    b2
        Tilt angle of azimuth axis, in East-West direction. East-positive.
    b3
        Elevation offset of encoder reading.
    g1
        Gravitational deflection coefficient.
    c1
    c2
    d1
    d2
    e1
    e2

    """

    def offset(self, az: u.Quantity, el: u.Quantity) -> Tuple[u.Quantity, u.Quantity]:
        dx = (
            self.a1 * np.sin(el)
            + self.a2
            + self.a3 * np.cos(el)
            + self.b1 * np.sin(az) * np.sin(el)
            - self.b2 * np.cos(az) * np.sin(el)
            + self.c1 * np.sin(az - el)
            + self.c2 * np.cos(az - el)
            + self.d1
            + self.e1 * np.cos(el)
            - self.e2 * np.sin(el)
        )
        dAz = dx / np.cos(el)
        dy = (
            self.b1 * np.cos(az)
            + self.b2 * np.sin(az)
            + self.b3
            + self.g1 * el
            + self.c1 * np.cos(az - el)
            - self.c2 * np.sin(az - el)
            + self.d2
            + self.e1 * np.sin(el)
            + self.e2 * np.cos(el)
        )
        dEl = dy

        # The above is defined as (refracted + offset = apparent), so reverse the sign
        return -1 * dAz, -1 * dEl

    def fit(self, *args, **kwargs):
        raise NotImplementedError("Fitting is not implemented for this model.")



from typing import Union

import astropy.units as u

from .signal_generator_base import SignalGenerator


class SignalGeneratorSimulator(SignalGenerator):
    Manufacturer: str = ""
    Model: str = ""
    Identifier = ""
    is_simulator = True

    def __init__(self) -> None:
        self.freq = 0
        self.power = 0
        self.output_status = False

    def set_freq(self, GHz: Union[int, float]) -> None:
        self.freq = GHz

    def set_power(self, dBm: Union[int, float]) -> None:
        self.power = dBm

    def get_freq(self) -> u.Quantity:
        return self.freq * u.GHz  # type: ignore

    def get_power(self) -> u.Quantity:
        return self.power * u.dBm  # type: ignore

    def start_output(self) -> None:
        self.output_status = True

    def stop_output(self) -> None:
        self.output_status = False

    def get_output_status(self) -> bool:
        return self.output_status

    def finalize(self) -> None:
        pass



"""Utility functions for arithmetic operations."""

__all__ = ["discretize", "counter", "ConditionChecker"]

import itertools
import math
from typing import Generator, Literal, Optional


def discretize(
    value: float,
    start: float = 0.0,
    step: float = 1.0,
    *,
    method: Literal["nearest", "ceil", "floor"] = "nearest",
) -> float:
    """Convert ``value`` to nearest element of arithmetic sequence.

    Parameters
    ----------
    value
        Parameter to discretize.
    start
        First element of element of arithmetic sequence.
    step
        Difference between the consecutive 2 elements of the sequence.
    method
        Discretizing method.

    Examples
    --------
    >>> neclib.utils.discretize(3.141592)
    3
    >>> neclib.utils.discretize(3.141592, step=10)
    0
    >>> neclib.utils.discretize(3.141592, method="ceil")
    4
    >>> neclib.utils.discretize(3.141592, start=2.5, step=0.7)
    3.2

    """
    discretizer = {"nearest": round, "ceil": math.ceil, "floor": math.floor}
    return discretizer[method]((value - start) / step) * step + start


def counter(
    stop: Optional[int] = None, allow_infty: bool = False
) -> Generator[int, None, None]:
    """Generate integers from 0 to ``stop``.

    Parameters
    ----------
    stop
        Number of yielded values.
    allow_infty
        If ``True``, the counter counts up to infinity. Listing such object will cause
        memory leak, so use caution.

    Examples
    --------
    >>> list(neclib.utils.counter(5))
    [0, 1, 2, 3, 4]
    >>> list(neclib.utils.counter())
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, ...]  # -> memory leak

    """
    if (stop is None) and (not allow_infty):
        raise ValueError("Specify ``stop`` value, unless ``allow_infty`` is set True.")
    elif stop is None:
        yield from itertools.count()
    elif stop < 0:
        raise ValueError("Stop value should be non-negative.")
    else:
        yield from range(stop)


class ConditionChecker:
    def __init__(self, sequential: int = 1, reset_on_failure: bool = True):
        self.__sequential = sequential
        self.__reset_on_failure = reset_on_failure
        self.__count = 0

    def check(self, condition: bool):
        self.__count += 1 if condition else 0
        if self.__reset_on_failure:
            self.__count *= int(condition)
        return self.__count >= self.__sequential



from typing import (
    Any,
    Callable,
    ClassVar,
    Generic,
    Literal,
    Protocol,
    Tuple,
    Type,
    TypeVar,
    Union,
    runtime_checkable,
)

import numpy as np
import numpy.typing as npt
from astropy.coordinates import BaseCoordinateFrame
from astropy.units import Quantity, UnitBase

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)


class Array(Protocol, Generic[T_co]):
    def __getitem__(self, index: Any, /) -> Union[T, "Array[T]"]:
        ...


DimensionLess = Union[int, float, npt.NDArray[np.number], Array[Union[int, float]]]
"""Type alias for values with no physical units."""

UnitType = Union[UnitBase, str]
"""Type alias for objects that represents physical unit."""

CoordFrameType = Union[str, BaseCoordinateFrame, Type[BaseCoordinateFrame]]
"""Type alias for objects that represents coordinate frame."""

EquivalencyType = Tuple[
    UnitBase, UnitBase, Callable[[float], float], Callable[[float], float]
]
"""Type alias for unit equivalency in ``astropy.units``."""

AngleUnit = Literal["deg", "rad", "arcmin", "arcsec"]
"""Type alias for supported angular units."""

CoordinateType = Tuple[Quantity, Quantity, CoordFrameType]
"""Type alias for coordinate in (lon, lat, frame) format."""


@runtime_checkable
class SupportsComparison(Protocol):
    def __eq__(self: T, other: T, /) -> bool:
        ...

    def __ne__(self: T, other: T, /) -> bool:
        ...

    def __lt__(self: T, other: T, /) -> bool:
        ...

    def __le__(self: T, other: T, /) -> bool:
        ...

    def __gt__(self: T, other: T, /) -> bool:
        ...


@runtime_checkable
class TextLike(Protocol):
    def upper(self) -> "TextLike":
        ...

    def lower(self) -> "TextLike":
        ...

    def find(self) -> int:
        ...

    def replace(self) -> "TextLike":
        ...

    def __len__(self) -> int:
        ...


@runtime_checkable
class IsDataClass(Protocol):
    __dataclass_fields__: ClassVar[dict]



from .da_converter_base import DAConverter


class DAConverterSimulator(DAConverter):
    Manufacturer: str = ""
    Model: str = ""
    Identifier = ""
    is_simulator = True

    def __init__(self) -> None:
        ...

    def set_voltage(self, mV: float, id: str) -> None:
        raise NotImplementedError

    def apply_voltage(self) -> None:
        raise NotImplementedError

    def finalize(self) -> None:
        pass



import math
from copy import deepcopy
from typing import Generic, List, Optional, Sequence, Tuple, TypeVar, Union

import numpy as np

T = TypeVar("T")
Index = TypeVar("Index", int, float)


class Interpolator(Generic[T]):
    """Interpolate Python objects.

    Example
    -------
    >>> interp = neclib.data.interpolate.Interpolator(align_by="x", attrs=["x", "y"])
    >>> x = SimpleNamespace(x=1, y=2)
    >>> xs = [SimpleNamespace(x=0, y=0), SimpleNamespace(x=2, y=100)]
    >>> interp(x, xs)
    SimpleNamespace(x=1.0, y=50.0)

    """

    def __init__(
        self, align_by: Optional[str] = None, attrs: Optional[Sequence[str]] = None
    ) -> None:
        self.align_by = align_by
        self.attrs = attrs

    def _get_sorted(self, x: T, xs: Sequence[T]) -> Tuple[Index, List[Index], List[T]]:
        if self.align_by is None:
            xs = sorted(xs)
            return x, xs, xs
        xs = sorted(xs, key=lambda _x: getattr(_x, self.align_by))
        return getattr(x, self.align_by), [getattr(_x, self.align_by) for _x in xs], xs

    def _get_value(self, lower: T, upper: T, ratio: Union[int, float]) -> T:
        if self.attrs is not None:
            for attr in self.attrs:
                try:
                    lower_attr, upper_attr = getattr(lower, attr), getattr(upper, attr)
                    interp_value = lower_attr + (upper_attr - lower_attr) * ratio
                    setattr(lower, attr, interp_value)
                except Exception:
                    pass
            return lower
        else:
            return lower + (upper - lower) * ratio

    def __call__(self, x: T, xs: Sequence[T]) -> T:
        raise NotImplementedError


class LinearInterp(Interpolator[T]):
    """Interpolate Python objects.

    Example
    -------
    >>> interp = neclib.data.interpolate.LinearInterp(align_by="x", attrs=["x", "y"])
    >>> x = SimpleNamespace(x=1, y=2)
    >>> xs = [SimpleNamespace(x=0, y=0), SimpleNamespace(x=2, y=100)]
    >>> interp(x, xs)
    SimpleNamespace(x=1.0, y=50.0)

    """

    def __call__(self, x: T, xs: Sequence[T]) -> T:
        _x, _xs, xs = self._get_sorted(x, xs)
        idx = np.interp(_x, _xs, range(len(_xs)))
        return self._get_value(
            deepcopy(xs[math.floor(idx)]), deepcopy(xs[math.ceil(idx)]), idx % 1
        )



"""Pure Python tools for NECST."""


# Perform time-consuming downloads
class _TimeConsumingTasks:
    @staticmethod
    def download_astropy_parameter_files():
        """Will download finals2000A.all (3.4MB) and Leap_Second.dat (1.3KB)."""
        from astropy.time import Time

        Time.now().ut1


import concurrent.futures  # noqa: E402

executor = concurrent.futures.ThreadPoolExecutor()
futures = [
    executor.submit(getattr(_TimeConsumingTasks, func))
    for func in dir(_TimeConsumingTasks)
    if not func.startswith("_")
]


# Logger configuration
import logging  # noqa: E402

rootLogger = logging.getLogger()
# Set minimum log level; not just for stream handler but for any handler attached.
# Stream handler will selectively handle logs of INFO or higher levels, but DEBUG level
# ones are not something you can completely ignore (may be recorded into log file)
rootLogger.setLevel(logging.DEBUG)
del logging, rootLogger


# Project version
from importlib.metadata import version  # noqa: E402

__version__ = version("neclib")
del version


# Subpackages
# `devices` isn't included, since they can be OS-dependent hence verbose warnings
from . import controllers  # noqa: F401, E402
from . import coordinates  # noqa: F401, E402
from . import core  # noqa: F401, E402
from . import recorders  # noqa: F401, E402
from . import safety  # noqa: F401, E402
from . import simulators  # noqa: F401, E402
from . import utils  # noqa: F401, E402

# Aliases
from .core import config, configure, get_logger  # noqa: F401, E402
from .core.data_type import *  # noqa: F401, E402, F403
from .core.exceptions import *  # noqa: F401, E402, F403

# Wait for all background tasks to complete. Timeout should be sufficiently large,
# otherwise attempt to download data files will be made every time importing `neclib`,
# until it completes in restricted duration.
concurrent.futures.wait(futures, timeout=None)
executor.shutdown()
del _TimeConsumingTasks, concurrent, executor, futures



__all__ = ["PointingError"]

import importlib
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Optional, Tuple, Union, overload

import astropy.units as u

from ...core import Parameters, get_quantity
from ...core.types import DimensionLess, UnitType


class PointingError(Parameters, ABC):
    """Calculate pointing error offset.

    Parameters
    ----------
    model
        Name of the pointing error model. If not specified, a dummy class which performs
        no pointing correction will be returned.
    **kwargs
        Model specific parameters.

    Notes
    -----
    This class automatically determines which pointing model to use based on the
    ``model`` argument, so you don't have to import observatory-specific subclass. The
    ``model`` can be specified in TOML file as well.

    Examples
    --------
    >>> pointing_error = neclib.parameters.PointingError.from_file(
    ...     "path/to/pointing_error.toml"
    ... )
    >>> pointing_error.apparent_to_refracted(0 * u.deg, 45 * u.deg)
    (<Quantity 0.1 deg>, <Quantity 45.5 deg>)
    >>> pointing_error.refracted_to_apparent(0 * u.deg, 45 * u.deg)
    (<Quantity -0.1 deg>, <Quantity 44.5 deg>)

    """

    @staticmethod
    def _normalize(key: str, /) -> str:
        return key.lower().replace("_", "")

    def __new__(cls, *, model: Optional[str] = None, **kwargs):
        if model is None:
            raise TypeError(
                f"Cannot instantiate abstract class {cls.__name__!r}. If you need a "
                "dummy version of this class, use `get_dummy` method."
            )

        model = cls._normalize(model)
        if model == cls._normalize(cls.__name__):
            return super().__new__(cls)

        subcls = {cls._normalize(v.__name__): v for v in cls.__subclasses__()}
        if model in subcls:
            return subcls[model](model=model, **kwargs)
        raise ValueError(
            f"Unknown pointing model: {model!r}\n"
            f"Supported ones are: {list(subcls.keys())}"
        )

    @classmethod
    def get_dummy(cls) -> "PointingError":
        """Return a dummy pointing error model which performs no correction.

        Examples
        --------
        >>> pointing_error = neclib.parameters.PointingError.get_dummy()
        >>> pointing_error.apparent_to_refracted(0 * u.deg, 45 * u.deg)
        (<Quantity 0. deg>, <Quantity 45. deg>)

        """

        class Dummy(PointingError):
            def fit(self, *args, **kwargs) -> Any:
                ...

            def offset(
                self, az: u.Quantity, el: u.Quantity
            ) -> Tuple[u.Quantity, u.Quantity]:
                return 0 * u.deg, 0 * u.deg  # type: ignore

        return Dummy(model="dummy")

    @abstractmethod
    def fit(self, *args, **kwargs) -> Any:
        """Fit the model to the measured pointing error parameters."""
        ...

    @abstractmethod
    def offset(self, az: u.Quantity, el: u.Quantity) -> Tuple[u.Quantity, u.Quantity]:
        """Compute the pointing error offset.

        Parameters
        ----------
        az
            Azimuth at which the pointing error is computed.
        el
            Elevation at which the pointing error is computed.

        Returns
        -------
        dAz
            Offset in azimuth axis.
        dEl
            Offset in elevation axis.

        Important
        ---------
        The offset will be ADDED to encoder readings to convert it to true sky/celestial
        coordinate.

        """
        ...

    @overload
    def apparent_to_refracted(
        self, az: u.Quantity, el: u.Quantity, unit: Optional[UnitType] = None
    ) -> Tuple[u.Quantity, u.Quantity]:
        ...

    @overload
    def apparent_to_refracted(
        self, az: DimensionLess, el: DimensionLess, unit: UnitType
    ) -> Tuple[u.Quantity, u.Quantity]:
        ...

    def apparent_to_refracted(
        self,
        az: Union[u.Quantity, DimensionLess],
        el: Union[u.Quantity, DimensionLess],
        unit: Optional[UnitType] = None,
    ) -> Tuple[u.Quantity, u.Quantity]:
        """Convert apparent AltAz coordinate to true coordinate.

        Parameters
        ----------
        az
            Apparent azimuth, which may not accurate due to pointing/instrumental error.
        el
            Apparent elevation, which may not accurate due to pointing/instrumental
            error.
        unit
            Unit of the input azimuth and elevation.

        Returns
        -------
        az
            True azimuth.
        el
            True elevation. Atmospheric refraction should be taken into account, when
            converting this to sky/celestial coordinate.

        Examples
        --------
        >>> pointing_error = neclib.parameters.PointingError.from_file(
        ...     "path/to/pointing_error.toml"
        ... )
        >>> pointing_error.apparent_to_refracted(0 * u.deg, 45 * u.deg)
        (<Quantity 0.1 deg>, <Quantity 45.5 deg>)

        """
        _az, _el = get_quantity(az, el, unit=unit)
        dAz, dEl = self.offset(_az, _el)
        return _az + dAz, _el + dEl

    @overload
    def refracted_to_apparent(
        self, az: u.Quantity, el: u.Quantity, unit: Optional[UnitType] = None
    ) -> Tuple[u.Quantity, u.Quantity]:
        ...

    @overload
    def refracted_to_apparent(
        self, az: DimensionLess, el: DimensionLess, unit: UnitType
    ) -> Tuple[u.Quantity, u.Quantity]:
        ...

    def refracted_to_apparent(
        self,
        az: Union[u.Quantity, DimensionLess],
        el: Union[u.Quantity, DimensionLess],
        unit: Optional[UnitType] = None,
    ) -> Tuple[u.Quantity, u.Quantity]:
        """Convert true sky/celestial coordinate to apparent AltAz coordinate.

        Parameters
        ----------
        az
            True azimuth.
        el
            True elevation. Atmospheric refraction should be taken into account before
            passing to this method.
        unit
            Unit of the input azimuth and elevation.

        Returns
        -------
        az
            Apparent azimuth, with pointing/instrumental error taken into account.
        el
            Apparent elevation, with pointing/instrumental error taken into account.

        Examples
        --------
        >>> pointing_error = neclib.parameters.PointingError.from_file(
        ...     "path/to/pointing_error.toml"
        ... )
        >>> pointing_error.refracted_to_apparent(0 * u.deg, 45 * u.deg)
        (<Quantity -0.1 deg>, <Quantity 44.5 deg>)

        """
        _az, _el = get_quantity(az, el, unit=unit)
        dAz, dEl = self.offset(_az, _el)
        return _az - dAz, _el - dEl


# Import all `PointingError` subclasses, to make them available in
# `PointingError.__subclasses__()` which is called in `PointingError.__new__()`.
impl = Path(__file__).parent.glob("*.py")
for p in impl:
    if p.name.startswith("_") or p.name == __file__:
        continue
    importlib.import_module(f"{__name__.rsplit('.', 1)[0]}.{p.stem}")



import numpy as np


class Functions:
    class Normal:
        def __init__(self, mu, sigma):
            self.mu = mu
            self.sigma = sigma

        def __call__(self, x):
            return (
                1
                / (self.sigma * np.sqrt(2 * np.pi))
                * np.exp(-((x - self.mu) ** 2) / (2 * self.sigma**2))
            )

    class Sigmoid:
        def __init__(self, mu, sigma):
            self.mu = mu
            self.sigma = sigma

        def __call__(self, x):
            return 1 / (1 + np.exp(-(x - self.mu) / self.sigma))



from functools import wraps


def disabled(func):
    """A decorator to disable a function or method.

    This may be used to make a tentative implementation not accessible.

    """

    @wraps(func)
    def _disabled(*args, **kwargs):
        raise NotImplementedError

    return _disabled
